package doxygen_csharp;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.digester.Digester;
import org.xml.sax.SAXException;

import com.change_vision.jude.api.inf.editor.BasicModelEditor;
import com.change_vision.jude.api.inf.editor.ModelEditorFactory;
import com.change_vision.jude.api.inf.editor.TransactionManager;
import com.change_vision.jude.api.inf.exception.InvalidEditingException;
import com.change_vision.jude.api.inf.exception.LicenseNotFoundException;
import com.change_vision.jude.api.inf.exception.ProjectLockedException;
import com.change_vision.jude.api.inf.exception.ProjectNotFoundException;
import com.change_vision.jude.api.inf.model.IModel;
import com.change_vision.jude.api.inf.project.ProjectAccessor;
import com.change_vision.jude.api.inf.project.ProjectAccessorFactory;


/**
 *	Parse Doxygen's xml to generate jude model 
 *
 */
public class DoxygenXmlParser {
	//for Installer
	private static final String FILE_NAME = "..\\..\\..\\template\\project\\C_Sharp.asta";
	//from Eclipse
//	private static final String FILE_NAME = "template\\project\\C_Sharp.asta";
	List compounds = new ArrayList();

	public void addCompound(Compound compound) {
		compounds.add(compound);
	}

	public static void main(String[] args) throws LicenseNotFoundException, ProjectLockedException, Throwable {
		// locate the doxygen xml's file
		String path = "";
		// locate generated jude-file
		String judePath = "";

		if (args.length == 2) {
			path = args[0];
			judePath = args[1];
		} else {
			print("Illegal Parameter! DoxygenXmlParser path_of_index.xml target_Astah_model.");
			System.exit(0);
		}

		File indexFile;
		if ((indexFile = findFile(path, "index.xml")) == null) {
			print("can't find index.xml.");
			System.exit(0);
		}

		//create a jude project
		ProjectAccessor prjAccessor = ProjectAccessorFactory.getProjectAccessor();
		// check template project
        File file = new File(FILE_NAME);
        boolean isUseTemplate = file.exists();
        if (isUseTemplate) {
        	prjAccessor.open(FILE_NAME);
        } else {
        	prjAccessor.create(judePath, false);
        }
		
		print("Importing... please wait..");
		
		//get current jude project's root
		IModel project = ProjectAccessorFactory.getProjectAccessor().getProject();

		DoxygenXmlParser doxygenXmlParser = new DoxygenXmlParser();

		//begin transaction
		TransactionManager.beginTransaction();
		
		//get the availability the concourse
		Set compounddefes = doxygenXmlParser.parserIndexXml(indexFile);

		for (Iterator iterator = compounddefes.iterator(); iterator.hasNext();) {
			//first deal namespace
			CompoundDef compounddef = (CompoundDef) iterator.next();
			if (CompoundDef.KIND_NAMESPACE.equals(compounddef.getCompounddefKind())||CompoundDef.KIND_INTERFACE.equals(compounddef.getCompounddefKind())) {
				(compounddef).convertToJudeModel(project, indexFile.getParentFile().listFiles());
			}
		}
		
		for (Iterator iterator = compounddefes.iterator(); iterator.hasNext();) {
			//second deal all the compound
			((CompoundDef) iterator.next()).convertToJudeModel(project, indexFile.getParentFile().listFiles());
		}
		
		for (Iterator iterator = compounddefes.iterator(); iterator.hasNext();) {
			//third convert all children
			((CompoundDef) iterator.next()).convertChildren(indexFile.getParentFile().listFiles());
		}
		//end transaction
		TransactionManager.endTransaction();
		
		//save the project
		if (isUseTemplate) {
			prjAccessor.saveAs(judePath);
		} else {
			prjAccessor.save();
		}
        
        //close the io
        prjAccessor.close();
        
        print("Import Done.");
	}
	
	/**
	 * 
	 * @param fileName: target indexFile
	 * @return the Set
	 * @throws IOException
	 * @throws SAXException
	 */
	private Set parserIndexXml(File indexFile) throws IOException, SAXException
			, ProjectNotFoundException, ClassNotFoundException, InvalidEditingException {
		// init digester for parser index.xml...
		Digester digester = new Digester();
		digester.setValidating(false);
		//create the DoxygenXmlParser
		digester.addObjectCreate("doxygenindex", DoxygenXmlParser.class);
		//create the Compound
		digester.addObjectCreate("doxygenindex/compound", Compound.class);
		digester.addSetProperties("doxygenindex/compound", "refid", "compoundRefid");
		digester.addSetProperties("doxygenindex/compound", "kind", "compoundKind");
		digester.addBeanPropertySetter("doxygenindex/compound/name", "name");

		digester.addSetNext("doxygenindex/compound", "addCompound");
		//init digester for parser  index.xml
		DoxygenXmlParser doxygenXmlParser = (DoxygenXmlParser) digester.parse(indexFile);
		
		for (Iterator iterator = doxygenXmlParser.compounds.iterator(); iterator.hasNext();) {
			Compound compound = (Compound) iterator.next();
			if ("file".equals(compound.getKind())) {
				//setLanguage
				String type = getFileType(compound.getName());
				LanguageManager.setCurrentLanguagePrimitiveType(type);
				
				//get current jude project's root
				IModel project = ProjectAccessorFactory.getProjectAccessor().getProject();
				
				BasicModelEditor basicModelEditor = ModelEditorFactory.getBasicModelEditor();
				//set the project type
				if (LanguageManager.LANGUAGE_CSHARP.equals(type)) {
					basicModelEditor.setLanguageCSharp(project, true);
				} else if (LanguageManager.LANGUAGE_JAVA.equals(type)) {
					basicModelEditor.setLanguageJava(project, true);
				}
			}
		}
		
		Set compounddefs = new HashSet();
		for (Iterator iterator = doxygenXmlParser.compounds.iterator(); iterator.hasNext();) {
			Compound compound = (Compound) iterator.next();
			//parse the kindName is class or interface or struct or union
			if ("namespace".equals(compound.getKind())
					|| "class".equals(compound.getKind())
					|| "interface".equals(compound.getKind())
					|| "struct".equals(compound.getKind())
					|| "union".equals(compound.getKind())
					|| "file".equals(compound.getKind())) {
				File file = findFile(indexFile.getParent(), compound.getRefid() + ".xml");
				compounddefs.add(parserCompounddefXML(file));
			}
		}
		return compounddefs;
	}
	
	/**
	 * 
	 * @param file: target file
	 * @return the Compounddef
	 * @throws IOException
	 * @throws SAXException
	 */
	public static CompoundDef parserCompounddefXML(File file) throws IOException, SAXException {
		// init digester for parser class.xml or interface.xml...
		Digester digester = new DoxygenDigester();
		digester.setValidating(false);
		//create the Compounddef
		digester.addRule("doxygen/compounddef", new ObjectCreateAfterLanguageRule(CompoundDef.class));
//		digester.addObjectCreate("doxygen/compounddef", Compounddef.class);
		digester.addSetProperties("doxygen/compounddef", "id", "compounddefId");
		digester.addSetProperties("doxygen/compounddef", "prot", "compounddefVisible");
		digester.addSetProperties("doxygen/compounddef", "kind", "compounddefKind");
		digester.addBeanPropertySetter("doxygen/compounddef/compoundname", "compoundName");
		
		//crreate the TempleParam
		digester.addRule("doxygen/compounddef/templateparamlist/param"
				, new ObjectCreateAfterLanguageRule(TemplateParam.class));
//		digester.addObjectCreate("doxygen/compounddef/templateparamlist/param", TempleParam.class);
		digester.addBeanPropertySetter("doxygen/compounddef/templateparamlist/param/type", "type");
		digester.addBeanPropertySetter("doxygen/compounddef/templateparamlist/param/declname", "declname");
		digester.addBeanPropertySetter("doxygen/compounddef/templateparamlist/param/defname", "defname");
		digester.addBeanPropertySetter("doxygen/compounddef/templateparamlist/param/defval", "defval");
		digester.addSetNext("doxygen/compounddef/templateparamlist/param", "addTemplateParam");
		
		//create the BaseCompounddefref
		digester.addRule("doxygen/compounddef/basecompoundref"
				, new ObjectCreateAfterLanguageRule(BaseCompoundDefRef.class));
//		digester.addObjectCreate("doxygen/compounddef/basecompoundref", BaseCompounddefref.class);
		digester.addSetProperties("doxygen/compounddef/basecompoundref", "refid", "refid");
		digester.addSetProperties("doxygen/compounddef/basecompoundref", "prot", "prot");
		digester.addSetProperties("doxygen/compounddef/basecompoundref", "virt", "virt");
		digester.addBeanPropertySetter("doxygen/compounddef/basecompoundref", "value");
		digester.addSetNext("doxygen/compounddef/basecompoundref", "addBasecompoundList");
		
		//create the Derivedcompoundref
		digester.addRule("doxygen/compounddef/derivedcompoundref"
				, new ObjectCreateAfterLanguageRule(DerivedCompoundRef.class));
//		digester.addObjectCreate("doxygen/compounddef/derivedcompoundref", Derivedcompoundref.class);
		digester.addSetProperties("doxygen/compounddef/derivedcompoundref", "refid", "refid");
		digester.addSetProperties("doxygen/compounddef/derivedcompoundref", "prot", "prot");
		digester.addSetProperties("doxygen/compounddef/derivedcompoundref", "virt", "virt");
		digester.addBeanPropertySetter("doxygen/compounddef/derivedcompoundref", "value");
		digester.addSetNext("doxygen/compounddef/derivedcompoundref", "addDerivedcompoundList");
		
		//create the InnerClass
		digester.addRule("doxygen/compounddef/innerclass"
				, new ObjectCreateAfterLanguageRule(InnerClass.class));
//		digester.addObjectCreate("doxygen/compounddef/innerclass", InnerClass.class);
		digester.addSetProperties("doxygen/compounddef/innerclass", "refid", "refid");
		digester.addSetProperties("doxygen/compounddef/innerclass", "prot", "prot");
		digester.addSetNext("doxygen/compounddef/innerclass", "addInnerclass");
		
		//create the Innernamespace
		digester.addRule("doxygen/compounddef/innernamespace"
				, new ObjectCreateAfterLanguageRule(InnerNameSpace.class));
//		digester.addObjectCreate("doxygen/compounddef/innernamespace", Innernamespace.class);
		digester.addSetProperties("doxygen/compounddef/innernamespace", "refid", "refid");
		digester.addSetNext("doxygen/compounddef/innernamespace", "addInnernamespace");
		
		digester.addSetProperties("doxygen/compounddef/location", "file", "locationFile");
		digester.addBeanPropertySetter("doxygen/compounddef/detaileddescription/para", "detaileddescriptionPara");
		
		//create the Section
		digester.addRule("doxygen/compounddef/sectiondef"
				, new ObjectCreateAfterLanguageRule(Section.class));
//		digester.addObjectCreate("doxygen/compounddef/sectiondef", Section.class);
		digester.addSetProperties("doxygen/compounddef/sectiondef", "kind", "kind");
		digester.addSetNext("doxygen/compounddef/sectiondef", "addSection");
		
		//create the Member
//		digester.addObjectCreate("doxygen/compounddef/sectiondef/memberdef", Member.class);
		digester.addRule("doxygen/compounddef/sectiondef/memberdef"
				, new ObjectCreateAfterLanguageRule(Member.class));
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "kind", "kind");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "prot", "prot");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "const", "constBoolean");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "static", "staticBoolean");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "gettable", "gettable");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "settable", "settable");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef", "virt", "virt");
		
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/type", "type");
		digester.addRule("doxygen/compounddef/sectiondef/memberdef/type/ref"
				, new ObjectCreateAfterLanguageRule(Ref.class));
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef/type/ref", "refid", "refid");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef/type/ref", "refid", "kindref");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/type/ref", "value");
		digester.addSetNext("doxygen/compounddef/sectiondef/memberdef/type/ref", "setTypeRef");

		//resolve special initializer
		digester.addRule("doxygen/compounddef/sectiondef/memberdef/initializer", new InitBeanPropertySetterRule());

		digester.addRule("doxygen/compounddef/sectiondef/memberdef/enumvalue"
				, new ObjectCreateAfterLanguageRule(EnumValue.class));
//		digester.addObjectCreate("doxygen/compounddef/sectiondef/memberdef/enumvalue", EnumValue.class);
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/enumvalue/name", "name");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/enumvalue/initializer", "initializer");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/enumvalue/briefdescription", "briefdescription");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/enumvalue/detaileddescription", "detaileddescription");
		digester.addSetNext("doxygen/compounddef/sectiondef/memberdef/enumvalue", "addEnum");

		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/name", "name");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/argsstring", "argsstring");
		
		//create the Param
		digester.addRule("doxygen/compounddef/sectiondef/memberdef/param"
				, new ObjectCreateAfterLanguageRule(Param.class));
//		digester.addObjectCreate("doxygen/compounddef/sectiondef/memberdef/param", Param.class);
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/param/type", "type");
		digester.addRule("doxygen/compounddef/sectiondef/memberdef/param/type/ref"
				, new ObjectCreateAfterLanguageRule(Ref.class));
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef/param/type/ref", "refid", "refid");
		digester.addSetProperties("doxygen/compounddef/sectiondef/memberdef/param/type/ref", "refid", "kindref");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/param/type/ref", "value");
		digester.addSetNext("doxygen/compounddef/sectiondef/memberdef/param/type/ref", "setTypeRef");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/param/declname", "declname");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/param/defname", "defname");
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/param/array", "array");
		digester.addSetNext("doxygen/compounddef/sectiondef/memberdef/param", "addMemberParam");
		
		digester.addBeanPropertySetter("doxygen/compounddef/sectiondef/memberdef/detaileddescription/para",
				"detaileddescriptionPara");
		digester.addSetNext("doxygen/compounddef/sectiondef/memberdef", "addMember");
		
		return (CompoundDef) digester.parse(file);
	}

	private static void print(String msg) {
		System.out.println(msg);
	}

	/**
	 * 
	 * @param path: target Path
	 * @param fileName: target file's name
	 * @return the target file
	 * @throws IOException
	 * @throws SAXException
	 */
	public static File findFile(String path, String fileName) throws IOException, SAXException {
		File dir = new File(path);
		if (dir.exists()) {
			File[] files = dir.listFiles();
			for (int i = 0; i < files.length; i++) {
				if (fileName.equals(files[i].getName())) {
					return files[i];
				}
			}
		}
		return null;
	}
	
	/**
	 * analyze fileName to judge Language. 
	 * *.cs: LANGUAGE_CSHARP
	 * *.java: LANGUAGE_JAVA
	 * *.cpp: LANGUAGE_CPLUS
	 * *.h: LANGUAGE_CPLUS
	 * *.cc: LANGUAGE_CPLUS
	 * @return language type
	 */

	public static String getFileType(String fileName) {
		String[] fileNames = fileName.split("\\.");

		if ("cs".equals(fileNames[fileNames.length - 1])) {
			return LanguageManager.LANGUAGE_CSHARP;
		}
		if ("java".equals(fileNames[fileNames.length - 1])) {
			return LanguageManager.LANGUAGE_JAVA;
		}
		if ("cpp".equals(fileNames[fileNames.length - 1])) {
			return LanguageManager.LANGUAGE_CPLUS;
		}
		if ("h".equals(fileNames[fileNames.length - 1])) {
			return LanguageManager.LANGUAGE_CPLUS;
		}
		if ("cc".equals(fileNames[fileNames.length - 1])) {
			return LanguageManager.LANGUAGE_CPLUS;
		}		
		return null;
	}
}